import functools
import numbers
import re
import unicodedata
from datetime import datetime
from time import sleep

import pandoc
import yaml
from pandoc.types import Meta, MetaBool, MetaList, MetaMap, MetaString, Space, Str
from plumbum import ProcessExecutionError

from n2y.errors import HTTPResponseError, PandocASTParseError
from n2y.logger import logger

# see https://pandoc.org/MANUAL.html#exit-codes
PANDOC_PARSE_ERROR = 64


def process_notion_date(notion_date):
    if notion_date is None:
        return None
    elif notion_date.get("end", None):
        return [
            notion_date["start"],
            notion_date["end"],
        ]
    else:
        return notion_date["start"]


def processed_date_to_plain_text(processed_date):
    # TODO: make this work the same way that Notion does
    if processed_date is None:
        return ""
    if isinstance(processed_date, list):
        return f"{processed_date[0]} to {processed_date[1]}"
    else:
        return processed_date


def pandoc_ast_to_markdown(pandoc_ast):
    # This function tries to avoid calling the separate pandoc binary (which is
    # slow) for basic cases with just spaces and strings
    if pandoc_ast is None or pandoc_ast == []:
        return ""
    elif type(pandoc_ast) is list and all(type(n) in [Str, Space] for n in pandoc_ast):
        # TODO: optimize performance for some other basic cases
        return "".join(" " if isinstance(n, Space) else n[0] for n in pandoc_ast)
    else:
        return pandoc_write_or_log_errors(
            pandoc_ast,
            format="markdown",
            options=[
                "--wrap",
                "none",  # don't hard line-wrap
                "--columns",
                "10000",  # The default column width is 72 characters.
                # When the width is this small, then pandoc may elect to generate HTML
                # tables in markdown instead of text-based tables; this is problematic
                # when we then convert the markdown into DOCX files which don't support raw HTML.
                "--eol",
                "lf",  # use linux-style line endings
            ],
        )


def pandoc_ast_to_html(pandoc_ast):
    return pandoc_write_or_log_errors(
        pandoc_ast,
        format="html+smart",
        options=[],
    )


def pandoc_write_or_log_errors(pandoc_ast, format, options):
    if pandoc_ast is None or pandoc_ast == []:
        return ""
    try:
        # TODO: add a mechanism to customize this
        return pandoc.write(pandoc_ast, format=format, options=options)
    except ProcessExecutionError as err:
        if err.retcode == PANDOC_PARSE_ERROR:
            lines = []
            # TODO: update this code to make it not print so much
            for element, path in pandoc.iter(pandoc_ast, path=True):
                path_str = ".".join(str(i) for _, i in path)
                lines.append(f"{path_str} {element}")
                logger.error("Pandoc AST:\n%s", "\n".join(lines))
            msg = (
                "Pandoc couldn't parse the generated AST. "
                f"This is likely due to a bug in n2y or a plugin: {err.stderr}"
            )
            logger.error(msg)
            raise PandocASTParseError(msg)
        else:
            raise


def yaml_to_meta_value(data):
    """
    Convert a Python object to a Pandoc metadata values, following the approach used by
    `yamlToMetaValue` here:

    https://github.com/jgm/pandoc/blob/main/src/Text/Pandoc/Readers/Metadata.hs.

    Note that all scalars end up as strings.
    """
    if isinstance(data, str):
        return MetaString(data)
    elif isinstance(data, bool):
        return MetaBool(data)
    elif data is None:
        return MetaString("")
    elif isinstance(data, numbers.Number):
        return MetaString(str(data))
    elif isinstance(data, list):
        return MetaList([yaml_to_meta_value(item) for item in data])
    elif isinstance(data, dict):
        return MetaMap({key: yaml_to_meta_value(value) for key, value in data.items()})
    else:
        logger.warning("Unsupported type %s for metadata value %s", type(data), data)
        return None


def yaml_map_to_meta(data):
    assert isinstance(data, dict)
    assert all(isinstance(k, str) for k in data.keys())
    return Meta({k: yaml_to_meta_value(v) for k, v in data.items()})


def pandoc_format_to_file_extension(format):
    # split on '-' or '+' to handle formats like 'markdown+raw_tex'
    base_type = re.split(r"[-+]", format)[0]
    if base_type in [
        "markdown",
        "gfm",
        "commonmark",
        "commonmark_x",
        "markdown_github",
        "markdown_mmd",
        "markdown_phpextra",
        "markdown_strict",
    ]:
        return "md"
    elif base_type in ["html", "html5", "html4"]:
        return "html"
    elif base_type in ["latex"]:
        return "tex"
    else:
        return base_type


def fromisoformat(datestring):
    """
    Parse Notion's datestrings, which aren't handled out of the box by
    `datetime.fromisoformat` because of the `Z` at the end of them.

    This function removes the need for a third-party library.
    """
    if datestring.endswith("Z"):
        return datetime.fromisoformat(datestring[:-1] + "+00:00")
    else:
        return datetime.fromisoformat(datestring)


def sanitize_filename(filename):
    """Taken from django."""
    s = str(filename).strip().replace(" ", "_")
    s = re.sub(r"(?u)[^-\w.]", "", s)
    if s in {".", ".."}:
        raise ValueError("Could not derive file name from '%s'" % filename)
    return s


def slugify(value, allow_unicode=False):
    """Taken from Django."""
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower())
    return re.sub(r"[-\s]+", "-", value).strip("-_")


def do_symbol(symbol, have_struck_letter):
    def do_special(potential_special):
        if (
            potential_special == "_"
            or potential_special == "-"
            or potential_special == "."
        ):
            return potential_special
        return ""

    def do_spacing(potential_spacing):
        if potential_spacing == " " or potential_spacing == "\n":
            return "-"
        return do_special(potential_spacing)

    def do_decimal(potential_decimal):
        if potential_decimal.isdecimal():
            return potential_decimal
        return do_spacing(potential_decimal)

    def do_letter(potential_letter):
        nonlocal have_struck_letter
        if potential_letter.isalpha():
            have_struck_letter = True
            return potential_letter.lower()
        if not have_struck_letter:
            return ""
        return do_decimal(potential_letter)

    return do_letter(symbol), have_struck_letter


def header_id_from_text(header_text, existing_ids=None):
    """
    Given the plain text for a header, produce a header.

    See https://pandoc.org/MANUAL.html#extension-auto_identifiers
    """
    have_struck_letter = False

    new_header_text = ""
    for symbol in header_text:
        symbol, have_struck_letter = do_symbol(symbol, have_struck_letter)
        new_header_text += symbol

    if len(new_header_text) == 0:
        new_header_text = "section"

    if existing_ids is not None:
        counter = 0
        duplicate_header_text = new_header_text
        while duplicate_header_text in existing_ids:
            counter += 1
            duplicate_header_text = f"{new_header_text}-{counter}"
        new_header_text = duplicate_header_text

    return new_header_text


def id_from_share_link(share_link):
    hyphens_removed = strip_hyphens(share_link)
    if not hyphens_removed.startswith("https://www.notion.so/"):
        return hyphens_removed
    else:
        domain_removed = hyphens_removed.split("/")[-1]
        query_removed = domain_removed.split("?")[0]
        assert len(query_removed) >= 32
        return query_removed[-32:]


def share_link_from_id(id):
    # Note that ordinarily page links include a hyphenated titled, but
    # fortunately they will redirect to the canonical page URL including the
    # hyphenated title if you visit the link with only the UUID. Similarly,
    # database urls often have a version parameter, but we can omit that too.
    return f"https://www.notion.so/{id}"


def strip_hyphens(string):
    return string.replace("-", "")


def stringify_list(array, wrap_in_quotes=False):
    if wrap_in_quotes:
        array = [f'"{item}"' for item in array]
    length = len(array)
    if length == 0:
        return ""
    elif length == 1:
        return array[0]
    elif length == 2:
        return f"{array[0]} and {array[1]}"
    else:
        return ", ".join(array[:-1]) + ", and " + array[-1]


def available_from_list(collection, singular, plural):
    if len(collection) == 0:
        return f"there are no available {plural}"
    elif len(collection) == 1:
        return f'the only available {singular} is "{collection[0]}"'
    else:
        return (
            f"the available {plural} are"
            f" {stringify_list(collection, wrap_in_quotes=True)}"
        )


def load_yaml(data):
    try:
        return yaml.load(data, Loader=yaml.SafeLoader)
    except yaml.YAMLError as e:
        raise ValueError('"{}" contains invalid YAML: {}'.format(data, e))


def retry_api_call(api_call):
    max_api_retries = 4

    @functools.wraps(api_call)
    def wrapper(*args, retry_count=0, **kwargs):
        assert "retry_count" not in kwargs, "retry_count is a reserved keyword"
        try:
            return api_call(*args, **kwargs)
        except HTTPResponseError as err:
            should_retry = err.status in [409, 429, 500, 502, 504, 503, 104]
            if not should_retry:
                raise err
            elif retry_count < max_api_retries:
                retry_count += 1
                if "retry-after" in err.headers:
                    retry_after = float(err.headers["retry-after"])
                    logger.info(
                        "This API call has been rate limited and "
                        "will be retried in %f seconds. Attempt %d of %d.",
                        retry_after,
                        retry_count,
                        max_api_retries,
                    )
                else:
                    retry_after = 2
                    logger.info(
                        "This API call failed and "
                        "will be retried in %f seconds. Attempt %d of %d.",
                        retry_after,
                        retry_count,
                        max_api_retries,
                    )
                sleep(retry_after)
                return wrapper(*args, retry_count=retry_count, **kwargs)
            else:
                raise err

    return wrapper



from n2y.notion import Client
from n2y.notion_mocks import mock_block, mock_rich_text
from tests.utils import NOTION_ACCESS_TOKEN


def test_append_and_delete_blocks():
    client = Client(NOTION_ACCESS_TOKEN, plugins=None)
    object_id = "c9e17a34da6a4b3295f82a1ad05bc3d8"
    page = client.get_page_or_database(object_id)
    notion_block = mock_block(
        "heading_2",
        {
            "color": "default",
            "rich_text": [mock_rich_text("m")],
        },
        has_children=True,
    )
    del notion_block["url"]
    creation_response = client.append_child_notion_blocks(
        page.notion_id, [notion_block]
    )
    new_block = creation_response[0]
    assert creation_response
    deletion_response = client.delete_notion_block(new_block)
    assert deletion_response


def test_append_child_page_or_database():
    """
    TODO: Add a short description of the purpose of this test.

    The destination location can be seen here:
    https://www.notion.so/Copy_To-Destination-c9e17a34da6a4b3295f82a1ad05bc3d8
    """
    client = Client(NOTION_ACCESS_TOKEN, plugins=None)
    destination_id = "c9e17a34da6a4b3295f82a1ad05bc3d8"
    original_id = "0b25a11e78b348c993b4dcf869f25a91"
    original = client.get_page_or_database(original_id)
    child_database = original.block.children[-1].notion_data
    child_page = original.block.children[-2].notion_data
    creation_response = client.append_child_notion_blocks(
        destination_id, [child_database, child_page]
    )
    assert creation_response
    for child in creation_response:
        deletion_response = client.delete_notion_block(child)
        assert deletion_response



import logging
import copy
import string

import yaml

from n2y.utils import strip_hyphens


logger = logging.getLogger(__name__)


DEFAULTS = {
    "media_root": "media",
    "media_url": "./media/",
}


EXPORT_DEFAULTS = {
    "yaml_front_matter": True,
    "id_property": "notion_id",
    "content_property": None,
    "url_property": "notion_url",
    "notion_filter": [],
    "notion_sorts": [],
    "pandoc_format": "markdown",
    "pandoc_options": [
        "--wrap",
        "none",  # don't hard line-wrap
        "--columns",
        "10000",  # The default column width is 72 characters.
        # When the width is this small, then pandoc may elect to generate HTML
        # tables in markdown instead of text-based tables; this is problematic
        # when we then convert the markdown into DOCX files which don't support raw HTML.
        "--eol",
        "lf",  # use linux-style line endings
    ],
    "plugins": [],
    "property_map": {},
    "filename_template": None,
}


def load_config(path):
    config = _load_config_from_yaml(path)
    if config is None:
        return None

    defaults_copy = copy.deepcopy(DEFAULTS)
    config = {**defaults_copy, **config}
    merged_exports = merge_config(
        config.get("exports", []),
        EXPORT_DEFAULTS,
        config.get("export_defaults", {}),
    )
    config["exports"] = merged_exports
    return config


def _load_config_from_yaml(path):
    try:
        with open(path, "r") as config_file:
            config = yaml.safe_load(config_file)
    except yaml.YAMLError as exc:
        logger.error("Error parsing the config file: %s", exc)
        return None
    except FileNotFoundError:
        logger.error("The config file '%s' does not exist", path)
        return None
    if not validate_config(config):
        logger.error("Invalid config file: %s", path)
        return None
    return config


def merge_config(config_items, builtin_defaults, defaults):
    """
    For each config item, merge in both the user provided defaults and the
    builtin defaults for each key value pair."
    """
    merged_config_items = []
    for config_item in config_items:
        master_defaults_copy = copy.deepcopy(builtin_defaults)
        defaults_copy = copy.deepcopy(defaults)
        config_item_copy = copy.deepcopy(config_item)
        merged_config_item = {
            **master_defaults_copy,
            **defaults_copy,
            **config_item_copy,
        }
        merged_config_items.append(merged_config_item)
    return merged_config_items


def merge_default_config(defaults):
    master_defaults_copy = copy.deepcopy(EXPORT_DEFAULTS)
    defaults_copy = copy.deepcopy(defaults)
    return {**master_defaults_copy, **defaults_copy}


def validate_config(config):
    if "exports" not in config:
        logger.error("Config missing the 'exports' key")
        return False
    if not isinstance(config["exports"], list) and len(config["exports"]) > 0:
        logger.error("Config 'exports' key must be a non-empty list")
        return False
    for export in config["exports"]:
        if not _validate_config_item(export):
            return False
    # TODO: validate the export defaults key
    return True


def valid_notion_id(notion_id):
    canonical_id = strip_hyphens(notion_id)
    return len(canonical_id) == 32 and canonical_id.isalnum()


def _validate_config_item(config_item):
    if "id" not in config_item:
        logger.error("Export config item missing the 'id' key")
        return False
    if not valid_notion_id(config_item["id"]):
        logger.error("Invalid id in export config item: %s", config_item["id"])
    if "node_type" not in config_item:
        logger.error("Export config item missing the 'node_type' key")
        return False
    if config_item["node_type"] not in [
        "page",
        "database_as_yaml",
        "database_as_files",
    ]:
        logger.error(
            "Invalid node_type in export config item: %s", config_item["node_type"]
        )
        return False
    if "filename_template" in config_item:
        if not _valid_filename_template(config_item["filename_template"]):
            return False
    if "output" not in config_item:
        logger.error("Export config item missing the 'output' key")
        return False
    if "notion_filter" in config_item:
        if not _valid_notion_filter(config_item["notion_filter"]):
            return False
    if "notion_sorts" in config_item:
        if not _valid_notion_sort(config_item["notion_sorts"]):
            return False
    # TODO: validate pandoc_format using the `--list-output-types` and `--list-extensions`
    # TODO: property map
    return True


def _valid_filename_template(filename_template):
    if not isinstance(filename_template, str):
        logger.error("filename_template must be a string")
        return False
    if filename_template == "":
        logger.error("filename_template must be a non-empty string")
        return False
    try:
        format_itr = string.Formatter().parse(filename_template)
        initial_format_tuple = next(format_itr)
    except ValueError as exc:
        logger.error("filename_template is invalid: %s", exc)
        return False
    initial_field_name = initial_format_tuple[1]
    if initial_field_name is None:
        logger.error("filename_template is invalid: Must be a format string")
        return False
    return True


def _valid_notion_filter(notion_filter):
    if not (isinstance(notion_filter, list) or isinstance(notion_filter, dict)):
        logger.error("notion_filter must be a list or dict")
        return False
    # TODO validate keys and values
    return True


def _valid_notion_sort(notion_sorts):
    if not (isinstance(notion_sorts, list) or isinstance(notion_sorts, dict)):
        logger.error("notion_sorts must be a list or dict")
        return False
    # TODO validate keys and values
    return True



from unittest.mock import patch

import pytest

from n2y.blocks import ParagraphBlock
from n2y.database import Database
from n2y.errors import PluginError
from n2y.mentions import PageMention
from n2y.notion import Client
from n2y.notion_mocks import (
    mock_database,
    mock_page,
    mock_page_mention,
    mock_paragraph_block,
    mock_user,
)
from n2y.page import Page
from n2y.plugins.dbfootnotes import PageMentionFootnote
from n2y.user import User


@patch("n2y.notion.Client.wrap_notion_user")
def mock_page_mention_with_footnote(
    mentioned_page_parent, wrap_notion_user, connect_parent_correctly=True
):
    client = Client("", plugins=["n2y.plugins.dbfootnotes"])
    wrap_notion_user.return_value = User(client, mock_user())
    # The original page, with the footnote ref.
    original_page = mock_page()
    # A paragraph in the original page where the mention occurs.
    paragraph_with_footnote_ref = mock_paragraph_block("Text with footnote ref")
    # The mention itself.
    mentioned_page = mock_page()
    mention = mock_page_mention()
    mention_text = "1"
    # Connect all of these mocked pages per the expected scenario.
    mention["page"]["id"] = mentioned_page["id"]
    # The parent of the mentioned page should be the inline DB.
    mentioned_page["parent"] = {
        "type": "page_id",
        "page_id": mentioned_page_parent["id"],
    }
    # The parent of the block with the mention should be the original page.
    paragraph_with_footnote_ref["parent"] = {
        "type": "page_id",
        "page_id": original_page["id"],
    }
    # In the correct scenario, the parent of the mentioned page's parent should
    # be the original page (the case of an inline DB).
    if connect_parent_correctly:
        mentioned_page_parent["parent"] = {
            "type": "page_id",
            "page_id": original_page["id"],
        }
        mentioned_page_parent_parent_n2y = Page(client, original_page)
    else:
        not_original_page = mock_page()
        mentioned_page_parent["parent"] = {
            "type": "page_id",
            "page_id": not_original_page["id"],
        }
        mentioned_page_parent_parent_n2y = Page(client, not_original_page)

    # Create `n2y` types from raw data constructed thus far.
    original_page_n2y = Page(client, original_page)
    mentioned_page_n2y = Page(client, mentioned_page)
    if mentioned_page_parent["object"] == "database":
        mentioned_page_parent_n2y = Database(client, mentioned_page_parent)
    else:
        mentioned_page_parent_n2y = Page(client, mentioned_page_parent)
    block = ParagraphBlock(client, paragraph_with_footnote_ref, original_page_n2y)

    page_mention_footnote_n2y = PageMentionFootnote.__new__(PageMentionFootnote)
    PageMention.__init__(page_mention_footnote_n2y, client, mention, mention_text)
    page_mention_footnote_n2y.mentioned_page = mentioned_page_n2y
    page_mention_footnote_n2y.block = block
    page_mention_footnote_n2y.mentioned_page_parent = mentioned_page_parent_n2y
    page_mention_footnote_n2y.mentioned_page_parent_parent = (
        mentioned_page_parent_parent_n2y
    )

    return page_mention_footnote_n2y


def test_check_that_mention_parent_is_a_database():
    from_page_mention_with_page_parent = mock_page_mention_with_footnote(mock_page())
    from_page_mention_with_db_parent = mock_page_mention_with_footnote(
        mock_database("Footnotes")
    )

    assert from_page_mention_with_page_parent._is_footnote() is False
    assert from_page_mention_with_db_parent._is_footnote() is True


def test_check_that_mention_parent_is_inline_db_of_original_page():
    from_page_mention_with_inline_db = mock_page_mention_with_footnote(
        mock_database("Footnotes"),
        connect_parent_correctly=True,
    )

    assert from_page_mention_with_inline_db._is_footnote() is True

    with pytest.raises(PluginError):
        mock_page_mention_with_footnote(
            mock_database("Footnotes"),
            connect_parent_correctly=False,
        )._is_footnote()


def test_footnote_db_naming_convention():
    page_mention_prefix = mock_page_mention_with_footnote(
        mock_database("Prefix Footnotes")
    )
    page_mention_suffix = mock_page_mention_with_footnote(
        mock_database("Footnotes Suffix")
    )
    page_mention_case_issue = mock_page_mention_with_footnote(
        mock_database("My footnotes")
    )
    page_mention_plural_issue = mock_page_mention_with_footnote(
        mock_database("My Footnote")
    )

    assert page_mention_prefix._is_footnote() is True
    assert page_mention_suffix._is_footnote() is False
    assert page_mention_case_issue._is_footnote() is False
    assert page_mention_plural_issue._is_footnote() is False



from pandoc.types import BulletList, Para, Str, Space, Header

from n2y.notion_mocks import mock_block, mock_rich_text, mock_paragraph_block
from tests.utils import block_colors
from tests.test_blocks import process_parent_block


def process_test_toggle_block(color):
    parent = mock_block(
        "toggle",
        {"rich_text": [mock_rich_text("Toggle text")], "color": color},
        has_children=True,
    )
    children = [
        mock_block("heading_1", {"rich_text": [mock_rich_text("Heading text")]}),
        mock_paragraph_block([("Paragraph text", [])]),
    ]
    return process_parent_block(parent, children, ["n2y.plugins.expandbluetoggles"])


def test_only_children_of_blue_toggles_are_rendered():
    pandoc_ast, markdown = process_test_toggle_block("blue_background")
    assert pandoc_ast == [
        Header(1, ("heading-text", [], []), [Str("Heading"), Space(), Str("text")]),
        Para([Str("Paragraph"), Space(), Str("text")]),
    ]
    assert (
        markdown
        == """# Heading text

Paragraph text
"""
    )


def test_non_blue_toggles_are_rendered_regularly_with_bullet_list():
    non_blue_colors = block_colors.copy()
    non_blue_colors.remove("blue")
    for color in non_blue_colors:
        pandoc_ast, markdown = process_test_toggle_block(color)
        pandoc_1 = BulletList(
            [
                [
                    Para([Str("Toggle"), Space(), Str("text")]),
                    Header(
                        1,
                        ("heading-text", [], []),
                        [Str("Heading"), Space(), Str("text")],
                    ),
                    Para([Str("Paragraph"), Space(), Str("text")]),
                ]
            ]
        )
        pandoc_2 = [
            Header(1, ("heading-text", [], []), [Str("Heading"), Space(), Str("text")]),
            Para([Str("Paragraph"), Space(), Str("text")]),
        ]
        markdown_1 = "# Heading text\n\nParagraph text\n"
        markdown_2 = "-   Toggle text\n\n    # Heading text\n\n    Paragraph text\n"
        assert pandoc_ast == pandoc_1 or pandoc_ast == pandoc_2
        assert markdown == markdown_1 or markdown == markdown_2



import hashlib
from os import makedirs, path
from urllib.parse import urlparse


from n2y.property_values import FilesPropertyValue
from n2y.utils import slugify


class DownloadFilePropertyValue(FilesPropertyValue):
    def to_value(self, _):
        url_list = []
        for file in self.files:
            file_content = self.client._get_url(file.url, stream=True)
            url_path = path.basename(urlparse(file.url).path)
            _, extension = path.splitext(url_path)
            file_hash = hashlib.sha256(file_content).hexdigest()[:10]
            page_title_slug = slugify(self.page.title.to_plain_text())
            file_name = f"{page_title_slug}-{file_hash}{extension}"
            full_filepath = path.join(self.client.media_root, file_name)
            makedirs(self.client.media_root, exist_ok=True)
            with open(full_filepath, "wb") as file:
                file.write(file_content)
            url_list.append(path.join(self.client.media_url, file_name))
        return url_list


notion_classes = {"property_values": {"files": DownloadFilePropertyValue}}



import os
import json
import tempfile
import subprocess
from pathlib import Path

from pandoc.types import Para, Image

from n2y.logger import logger
from n2y.errors import UseNextClass
from n2y.blocks import FencedCodeBlock

mermaid_config = {"flowchart": {"useMaxWidth": False}}

puppeteer_config = {
    "headless": True,
    "args": [
        "--no-sandbox",
        "--disable-setuid-sandbox",
    ],
}


class MermaidFencedCodeBlock(FencedCodeBlock):
    """
    Adds support for generating mermaid diagrams from codeblocks with the
    "mermaid" language, as supported in the Notion UI.

    This plugin assumes that the `mmdc` mermaid commandline tool is available,
    and will throw an exception if it is not.

    If there are errors with the mermaid syntax, it is treated as a normal
    codeblock and the warning is logged.
    """

    def __init__(self, client, page, notion_data, get_children=True):
        super().__init__(client, page, notion_data, get_children)
        if self.language != "mermaid":
            raise UseNextClass()

    def to_pandoc(self):
        # TODO: Clean up by extracting all this temp code out
        temp_fd, temp_filepath = tempfile.mkstemp(suffix=".png")
        os.close(temp_fd)
        temp_config_mermaid_fd, temp_config_mermaid_filepath = tempfile.mkstemp(
            suffix=".json"
        )
        os.write(temp_config_mermaid_fd, json.dumps(mermaid_config).encode("utf-8"))
        os.close(temp_config_mermaid_fd)
        temp_config_puppeteer_fd, temp_config_puppeteer_filepath = tempfile.mkstemp(
            suffix=".json"
        )
        os.write(temp_config_puppeteer_fd, json.dumps(puppeteer_config).encode("utf-8"))
        os.close(temp_config_puppeteer_fd)
        try:
            diagram_as_text = self.rich_text.to_plain_text()
            diagram_as_bytes = diagram_as_text.encode()
            subprocess.run(
                [
                    "mmdc",
                    "--configFile",
                    temp_config_mermaid_filepath,
                    "--puppeteerConfigFile",
                    temp_config_puppeteer_filepath,
                    "-o",
                    temp_filepath,
                ],
                capture_output=True,
                input=diagram_as_bytes,
                check=True,
            )
            with open(temp_filepath, "rb") as temp_file:
                content = temp_file.read()
                root = Path(__file__).resolve().parent.parent
                with open(root / "data" / "mermaid_err.png", "rb") as err_img:
                    if content == err_img.read():
                        raise NotImplementedError("Syntax Error In Graph")
                url = self.client.save_file(content, self.page, ".png", self.notion_id)
                caption = []
                fig_flag = ""
                if self.caption:
                    fig_flag = "fig:"
                    caption = self.caption.to_pandoc()
            return Para([Image(("", [], []), caption, (url, fig_flag))])
        except subprocess.CalledProcessError as exc:
            # as of now, mmdc does not ever return a non-zero error code, so
            # this won't ever be hit
            msg = (
                "Unable to convert mermaid diagram (%s) into an image. "
                "The mermaid-cli returned error code %d and printed: %s"
            )
            logger.error(msg, self.notion_url, exc.returncode, exc.stderr)
        except subprocess.SubprocessError:
            msg = "Unable to convert mermaid diagram (%s) into an image"
            logger.exception(msg, self.notion_url)
        except NotImplementedError:
            msg = (
                "Unable to convert mermaid diagram (%s) into"
                " an image due to a syntax error in the graph"
            )
            logger.exception(msg, self.notion_url)
        except FileNotFoundError:
            msg = (
                "Unable to find the mermaid-cli executable, `mmdc`, on the PATH. "
                "See here: https://github.com/mermaid-js/mermaid-cli "
                "Mermaid diagram (%s) in code blocks will not be converted to images."
            )
            logger.error(msg, self.notion_url)
        return super().to_pandoc()


notion_classes = {
    "blocks": {
        "code": MermaidFencedCodeBlock,
    }
}



from n2y.blocks import HeadingBlock

from pandoc.types import Header, Link


class LinkedHeadingBlock(HeadingBlock):
    """
    Make headers have links back to the originating notion block.
    """

    def to_pandoc(self):
        link = [Link(("", [], []), self.rich_text.to_pandoc(), (self.notion_url, ""))]
        return Header(self.level, ("", [], []), link)


class LinkedHeadingThreeBlock(LinkedHeadingBlock):
    level = 3


class LinkedHeadingTwoBlock(LinkedHeadingBlock):
    level = 2


class LinkedHeadingOneBlock(LinkedHeadingBlock):
    level = 1


notion_classes = {
    "blocks": {
        "heading_1": LinkedHeadingOneBlock,
        "heading_2": LinkedHeadingTwoBlock,
        "heading_3": LinkedHeadingThreeBlock,
    }
}



"""
The JinjaRenderPage plugin makes it possible to use Jinja2 templates to render
content from Notion databases into the Pandoc AST.

In addition to the raw power of Jinja2, this plugin provides a few useful Jinja
filters and context variables. One variable in particular, the FirstPassOutput
object, is used to make it possible to render content in two passes. This is
useful for rendering content that depends on other content in the same page. For
example, if one ones to generate a glossary from a centralized "Terms" database
and to only include terms which are present on the page, one can use the
FirstPassOutput filter to first generate the content of the page, and then make
a second pass to render the glossary.

Since the FirstPassOutput object requires a full render of the page, this plugin
overrides the Page class.

In addition, it also overrides the CodeBlock class.

Any code block whose caption begins with "{jinja=pandoc_format}" will be rendered
using Jinja2 into text, which in turn will be parsed by Pandoc into assuming it
follows the pandoc_format, into the AST. Thus, the jinja can be used to render
latex, html, or any other format that Pandoc supports into the AST.

Another context variable that is provided is the database variable, which is
used to access Notion databases. Any database that is @-mentioned in the caption
of the codeblock may be accessed via the database name.
"""
import functools
import re
import traceback

import jinja2
import pandoc
from pandoc.types import Code, Meta, MetaString, Pandoc, Para, Plain, Str

from n2y.blocks import FencedCodeBlock, HeadingBlock
from n2y.errors import UseNextClass
from n2y.export import database_to_yaml
from n2y.logger import logger
from n2y.mentions import DatabaseMention
from n2y.page import Page
from n2y.rich_text import MentionRichText
from n2y.utils import available_from_list, pandoc_ast_to_markdown


def join_to(foreign_keys, table, primary_key="notion_id"):
    """
    Given a set of ids for an object, and a list of the objects these ids refer
    to, select out the objects by joining using the specified primary key
    (which defaults to 'id').
    """
    joined = []
    for foreign_key in foreign_keys:
        selected_row = None
        for row in table:
            if row[primary_key] == foreign_key:
                selected_row = row
                break
        joined.append(selected_row)
    return joined


def list_matches(string, text):
    return list(
        re.finditer(
            "(?<![a-zA-Z])"
            + re.escape(_canonicalize(string))
            + "(?:s|es)?(?![a-zA-Z])",
            _canonicalize(text),
            re.IGNORECASE,
        )
    )


def remove_words(words, text):
    for word in words:
        first, last = word.span()
        text = text[:first] + " " * (last - first) + text[last:]
    return text


def _fuzzy_find_in(term_list, text, key="Name", by_length=True, reverse=True):
    found = []
    key_filter = lambda d: len(d[key]) if by_length else d[key]
    sorted_term_list = sorted(term_list, key=key_filter, reverse=reverse)
    if key in term_list[0]:
        for term in sorted_term_list:
            if key in term and term[key] != "":
                matches = list_matches(term[key], text)
                if matches != []:
                    found.append(term)
                    text = remove_words(matches, text)
    return found


def fuzzy_find_in(term_list, text, key="Name", by_length=True, reverse=True):
    """
    Used to search a markdown string, which may have been modified using pandoc's smart extension,
    for the value at a specified key in a dicionary for all dictionaries in a given list of them.

    see https://pandoc.org/MANUAL.html#extension-smart
    """
    found = []

    if isinstance(key, str):
        found = _fuzzy_find_in(term_list, text, key, by_length, reverse)
    elif isinstance(key, list):
        keys = key
        for key in keys:
            terms_found = fuzzy_find_in(term_list, text, key, by_length, reverse)
            found.extend(terms_found)

    return found


def _canonicalize(markdown):
    markdown = markdown.replace("\u201D", '"').replace("\u201C", '"')
    markdown = markdown.replace("\u2019", "'").replace("\u2018", "'")
    markdown = markdown.replace("\u2013", "--").replace("\u2014", "---")
    markdown = markdown.replace("\u2026", "...")
    markdown = markdown.replace("\u00A0", " ")
    return markdown


def _create_jinja_environment():
    environment = jinja2.Environment(
        cache_size=0,
        undefined=jinja2.StrictUndefined,
        extensions=["jinja2.ext.do"],
    )
    environment.globals["first_pass_output"] = FirstPassOutput()
    environment.filters["fuzzy_find_in"] = fuzzy_find_in
    environment.filters["join_to"] = join_to
    return environment


def render_from_string(source, context=None, environment=None):
    if environment is None:
        environment: jinja2.Environment = _create_jinja_environment()
    if context is None:
        context = {}
    template = environment.from_string(source)
    output = template.render(context)

    # output string usually loses trailing new line.
    if output and output[-1] != "\n":
        output += "\n"
    return output


class JinjaExceptionInfo:
    err: Exception
    object: object
    method: str
    args: list
    kwargs: dict | None

    def __init__(self, object, err, method, args, kwargs):
        self.object = object
        self.method = method
        self.kwargs = kwargs
        self.args = args
        self.err = err


class JinjaCacheObject(dict):
    def __init__(self, *args, block=None, **kwargs):
        if block is None:
            raise ValueError(
                f"block must be specified when instantiating {self.__class__}."
            )
        self.block = block
        super().__init__(*args, **kwargs)

    def __getitem__(self, __key):
        try:
            return super().__getitem__(__key)
        except Exception as err:
            self.cache_exception_info(err, "__getitem__", [__key])
            raise

    def cache_exception_info(self, *args, kwargs=None):
        self.block.exc_info = JinjaExceptionInfo(self, *args, kwargs)


class JinjaDatabaseItem(JinjaCacheObject):
    ...


class PageProperties(JinjaCacheObject):
    ...


class JinjaDatabaseCache(JinjaCacheObject):
    def __getitem__(self, __key):
        try:
            if isinstance(__key, str):
                return super().__getitem__(__key)
            elif isinstance(__key, int):
                return super().__getitem__(list(self)[__key])
        except Exception as err:
            self.cache_exception_info(err, "__getitem__", [__key])
            raise
        logger.error(
            "Jinja template database must be selected using a string or an integer"
        )
        err = KeyError(__key)
        self.cache_exception_info(err, "__getitem__", [__key])
        raise err

    def __setitem__(self, __key: str, __value: list):
        return super().__setitem__(
            __key, [JinjaDatabaseItem(item, block=self.block) for item in __value]
        )


class FirstPassOutput:
    def __init__(self, lines=None):
        self._second_pass_is_requested = False
        self._lines = lines
        self._source = None

    def __bool__(self):
        return bool(self._lines)

    def request_second_pass(self):
        self._second_pass_is_requested = True

    @property
    def is_second_pass(self):
        return self._lines is not None

    @property
    def second_pass_is_requested(self):
        return not self.is_second_pass and self._second_pass_is_requested

    @property
    def lines(self):
        self.is_second_pass or self.request_second_pass()
        return self._lines or []

    def set_lines(self, lines):
        self._lines = lines

    @property
    def source(self):
        if self._source is None or self.is_second_pass:
            self._source = "\n".join(self.lines)
        return self._source


class JinjaRenderPage(Page):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        if "jinjarenderpage" not in client.plugin_data:
            client.plugin_data["jinjarenderpage"] = {}
        if self.notion_id not in client.plugin_data["jinjarenderpage"]:
            self.jinja_environment = _create_jinja_environment()
            client.plugin_data["jinjarenderpage"][self.notion_id] = {
                "environment": self.jinja_environment,
            }
        else:
            self.jinja_environment = client.plugin_data["jinjarenderpage"][
                self.notion_id
            ]["environment"]

    def to_pandoc(self, ignore_toc=False):
        ast = super().to_pandoc(ignore_toc=True)
        first_pass_output = self.jinja_environment.globals["first_pass_output"]
        if first_pass_output.second_pass_is_requested:
            first_pass_output_text = pandoc_ast_to_markdown(ast)
            first_pass_output.set_lines(
                first_pass_output_text.splitlines(keepends=True)
            )
            ast = super().to_pandoc(ignore_toc=True)
            jinja2.clear_caches()
        return ast if ignore_toc else self.generate_toc(ast)


class JinjaFencedCodeBlock(FencedCodeBlock):
    trigger_regex = re.compile(r"^{jinja=(.+)}")

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        result = self.caption.matches(self.trigger_regex)
        if result:
            self.pandoc_format: str = result.group(1)
        else:
            raise UseNextClass()
        self.render_count: int = 0
        self.rendered_text: str = ""
        self.error: str | None = None
        self.exc_info: JinjaExceptionInfo | None = None
        self.databases: JinjaDatabaseCache | None = None
        self.jinja_code: str = self.rich_text.to_plain_text()
        self.uses_first_pass_output: bool = "first_pass_output" in self.jinja_code
        self.page_props: PageProperties = self.page.properties_to_values(
            self.pandoc_format
        )
        self.jinja_environment: jinja2.Environment = self.client.plugin_data[
            "jinjarenderpage"
        ][self.page.notion_id]["environment"]

    def _get_database_ids_from_mentions(self):
        for rich_text in self.caption:
            is_mention = isinstance(rich_text, MentionRichText)
            if is_mention and isinstance(rich_text.mention, DatabaseMention):
                yield rich_text.mention.notion_database_id

    def _get_yaml_from_mentions(self):
        self.databases = JinjaDatabaseCache(block=self)
        export_defaults = self.client.export_defaults
        for database_id in self._get_database_ids_from_mentions():
            database = self.client.get_database(database_id)
            # TODO: Rethink about the database data is accessed from within the
            # templates; perhaps it should be something more like Django's ORM
            # where we can filter and sort the databases via the API, instead of
            # having to pull in ALL of the data first before filtering it.
            # Such an API would also alleviate the need to pass in the export
            # defaults and it could provide a mechanism to access the page
            # content from within the Jinja templates, which isn't possible
            # right now.
            if (database_name := database.title.to_plain_text()) in self.databases:
                msg = (
                    f'Duplicate database name "{database_name}"'
                    f" when rendering [{self.notion_url}]"
                )
                logger.error(msg)
                raise ValueError(msg)
            self.databases[database_name] = database_to_yaml(
                database=database,
                pandoc_format=self.pandoc_format,
                pandoc_options=[],
                id_property=export_defaults["id_property"],
                url_property=export_defaults["url_property"],
            )

    def _specify_err_msg(self, err: Exception):
        block_ref: str = f"See the Notion code block here: {self.notion_url}."
        line_num: str = traceback.format_exc().split('>", line ')[1][0]

        if self.exc_info is not None:
            if type(self.exc_info.err) is KeyError:
                if type(self.exc_info.object) is JinjaDatabaseCache:
                    available_props: str = available_from_list(
                        list(self.exc_info.object), "database", "databases"
                    )
                    return (
                        f' You attempted to access the "{self.exc_info.args[0]}" database'
                        f" on line {line_num} of said template, but {available_props}."
                        " Note that databases must be mentioned in the Notion code"
                        " block's caption to be available and the plugin must have"
                        " permission to read the database via the NOTION_ACCESS_TOKEN."
                        f" {block_ref}"
                    )
                elif type(self.exc_info.object) is JinjaDatabaseItem:
                    available_props: str = available_from_list(
                        list(self.exc_info.object), "property", "properties"
                    )
                    return (
                        f' You attempted to access the "{self.exc_info.args[0]}" property'
                        f" of a database item on line {line_num} of said template, but"
                        f" {available_props}. {block_ref}"
                    )
                elif type(self.exc_info.object) is PageProperties:
                    available_props: str = available_from_list(
                        list(self.exc_info.object), "property", "properties"
                    )
                    return (
                        f' You attempted to access the "{self.exc_info.args[0]}" property'
                        f" of this page on line {line_num} of said template, but"
                        f" {available_props}. {block_ref}"
                    )
            elif self.exc_info.object in ["test", "filter"]:
                return (
                    f' Recieved the message "{str(err)}" when evaluating line {line_num}.'
                    f' The Jinja {self.exc_info.object} "{self.exc_info.method}" raised'
                    " this error when called with the following argument(s):"
                    f' {{\n\t"args": {self.exc_info.args},\n\t"kwargs":'
                    f" {self.exc_info.kwargs}\n}}\n{block_ref}"
                )
        elif type(err) is jinja2.exceptions.TemplateSyntaxError:
            return (
                f" There is a syntax error in the Jinja template on line {line_num}."
                f" Recieved the message {str(err)}. {block_ref}"
            )

        return (
            f' Received the message "{str(err)}" when evaluating line {line_num}.'
            f" {block_ref}"
        )

    def _log_jinja_error(self, err: Exception):
        self.error = (
            f"Error rendering a Jinja template on {self.page.title.to_plain_text()}."
            if self.page
            else "Unknown Page."
        )
        self.error += self._specify_err_msg(err)
        logger.error(self.error)

    def _render_error(self, err, during_render=True):
        if self.render_count == 1 or during_render and self.render_count == 0:
            self._log_jinja_error(err)

    def _error_ast(self):
        return [Para([Code(("", ["markdown"], []), self.error)])]

    def _render_text(self):
        self.jinja_environment.filters = {
            k: self._debug_assist(v, "filter", k)
            for k, v in self.jinja_environment.filters.items()
        }
        self.jinja_environment.tests = {
            k: self._debug_assist(v, "test", k)
            for k, v in self.jinja_environment.tests.items()
        }
        if not getattr(self, "context", None):
            self.context = {
                "databases": self.databases,
                "page": PageProperties(self.page_props, block=self),
            }
        if "render_content" in self.jinja_code:

            def render_content(notion_id, level_adjustment=0):
                page = self.client.get_page(notion_id)
                for child in page.block.children:
                    if isinstance(child, HeadingBlock):
                        child.level = max(1, child.level + level_adjustment)
                ast = (
                    page.to_pandoc()
                    if page.to_pandoc()
                    else Pandoc(Meta({"title": MetaString(page.block.title)}), [])
                )
                content = pandoc.write(
                    ast,
                    format=self.pandoc_format,
                    options=[],
                )
                return content

            self.jinja_environment.filters["render_content"] = self._debug_assist(
                render_content, "filter", "render_content"
            )
        try:
            self.rendered_text = render_from_string(
                self.jinja_code, self.context, self.jinja_environment
            )
        except Exception as err:
            self._render_error(err)
        self.render_count += 1

    def to_pandoc(self):
        if self.databases is None:
            self._get_yaml_from_mentions()
        if (
            self.render_count < 1
            or self.render_count < 2
            and self.uses_first_pass_output
        ):
            self._render_text()
        if self.error:
            children_ast = self._error_ast()
        else:
            if self.pandoc_format != "plain":
                # pandoc.read includes Meta data, which isn't relevant here; we just
                # want the AST for the content
                try:
                    document_ast = pandoc.read(
                        self.rendered_text, format=self.pandoc_format
                    )
                    children_ast = document_ast[1]
                except Exception as err:
                    self._render_error(err, during_render=False)
                    children_ast = self._error_ast()
            else:
                # Pandoc doesn't support reading "plain" text into it's AST (since
                # if it was just plain text, why would you need pandoc to parse it!)
                # That said, sometimes it is useful to create plain text from the
                # jinja rendering (e.g., when producing a site map or something
                # similar from Notion databases).
                children_ast = Plain([Str(self.rendered_text)])
        return children_ast

    def _debug_assist(self, func, jinja_type, jinja_ref):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as err:
                if self.exc_info is None:
                    if type(args[0]) is jinja2.Environment:
                        args = list(args[1:])
                    self.exc_info = JinjaExceptionInfo(
                        jinja_type, err, jinja_ref, args, kwargs
                    )
                raise

        return wrapper


notion_classes = {
    "page": JinjaRenderPage,
    "blocks": {
        "code": JinjaFencedCodeBlock,
    },
}



from pandoc.types import Str, Space, LineBreak

from n2y.rich_text import RichText


def test_plain_text_to_pandoc_basic():
    pandoc_ast = RichText.plain_text_to_pandoc("hello world")
    assert pandoc_ast == [Str("hello"), Space(), Str("world")]


def test_plain_text_to_pandoc_spaces_before():
    pandoc_ast = RichText.plain_text_to_pandoc("  hello")
    assert pandoc_ast == [Space(), Space(), Str("hello")]


def test_plain_text_to_pandoc_spaces_after():
    pandoc_ast = RichText.plain_text_to_pandoc("hello ")
    assert pandoc_ast == [Str("hello"), Space()]


def test_plain_text_to_pandoc_spaces_after_newline():
    pandoc_ast = RichText.plain_text_to_pandoc("hello\n  world")
    assert pandoc_ast == [Str("hello"), LineBreak(), Space(), Space(), Str("world")]



from unittest.mock import patch

import pytest

from n2y import notion
from n2y.notion_mocks import (
    mock_formula_property_value,
    mock_person_user,
    mock_property_value,
    mock_relation_value,
    mock_rich_text,
    mock_rich_text_array,
    mock_rollup_property_value,
    mock_select_option,
    mock_user,
)
from n2y.user import User


@patch("n2y.notion.Client.wrap_notion_user")
def process_property_value(notion_data, wrap_notion_user):
    client = notion.Client("")
    if notion_data.get("people", None):
        wrap_notion_user.return_value = User(client, notion_data["people"][0])
    else:
        wrap_notion_user.return_value = User(client, mock_user())
    property_value = client.wrap_notion_property_value(notion_data, None)
    return property_value.to_value("gfm")


def test_title():
    notion_data = mock_property_value("title", [mock_rich_text("text")])
    assert process_property_value(notion_data) == "text"


def test_rich_text_empty():
    notion_data = mock_property_value("rich_text", [])
    assert process_property_value(notion_data) == ""


def test_rich_text_annotated():
    notion_data = mock_property_value(
        "rich_text",
        mock_rich_text_array([("Hello", ["bold"]), (" ", []), ("Goodbye", "italic")]),
    )
    assert process_property_value(notion_data) == "**Hello** *Goodbye*\n"


def test_number_empty():
    notion_data = mock_property_value("number", None)
    assert process_property_value(notion_data) is None


def test_number_non_empty():
    notion_data = mock_property_value("number", 3)
    assert process_property_value(notion_data) == 3


def test_status():
    notion_data = mock_property_value("status", "red")
    assert process_property_value(notion_data) == "red"


def test_select_empty():
    notion_data = mock_property_value("select", None)
    assert process_property_value(notion_data) is None


def test_select_non_empty():
    notion_data = mock_property_value("select", mock_select_option("A"))
    assert process_property_value(notion_data) == "A"


def test_multi_select_empty():
    notion_data = mock_property_value("multi_select", [])
    assert process_property_value(notion_data) == []


def test_multi_select_non_empty():
    notion_data = mock_property_value(
        "multi_select",
        [
            mock_select_option("A"),
            mock_select_option("B"),
        ],
    )
    assert process_property_value(notion_data) == ["A", "B"]


def test_date_empty():
    notion_data = mock_property_value("date", None)
    assert process_property_value(notion_data) is None


def test_date_start_only():
    notion_data = mock_property_value("date", {"start": "2022-05-11"})
    assert process_property_value(notion_data) == "2022-05-11"


def test_date_start_and_end():
    notion_data = mock_property_value(
        "date", {"start": "2022-05-11", "end": "2022-05-12"}
    )
    assert process_property_value(notion_data) == ["2022-05-11", "2022-05-12"]


def test_people_empty():
    notion_data = mock_property_value("people", [])
    assert process_property_value(notion_data) == []


def test_people_minimal_user():
    notion_data = mock_property_value("people", [mock_user()])
    assert process_property_value(notion_data) == [""]


def test_people_person_user():
    notion_person = mock_person_user("Happy Man", "happyman@gmail.com")
    notion_data = mock_property_value("people", [notion_person])
    assert process_property_value(notion_data) == ["Happy Man"]


# TODO: add tests for bot users


def test_checkbox():
    notion_data = mock_property_value("checkbox", True)
    assert process_property_value(notion_data)


def test_url_empty():
    notion_data = mock_property_value("url", None)
    assert process_property_value(notion_data) is None


def test_url_non_empty():
    notion_data = mock_property_value("url", "https://innolitics.com")
    assert process_property_value(notion_data) == "https://innolitics.com"


def test_phone_number_empty():
    notion_data = mock_property_value("phone_number", None)
    assert process_property_value(notion_data) is None


def test_phone_number_non_empty():
    notion_data = mock_property_value("phone_number", "1-555-555-5555")
    assert process_property_value(notion_data) == "1-555-555-5555"


def test_email_empty():
    notion_data = mock_property_value("email", None)
    assert process_property_value(notion_data) is None


def test_email_non_empty():
    notion_data = mock_property_value("email", "info@innolitics.com")
    assert process_property_value(notion_data) == "info@innolitics.com"


def test_formula_number_empty():
    notion_data = mock_formula_property_value("number", None)
    assert process_property_value(notion_data) is None


def test_formula_number_non_empty():
    notion_data = mock_formula_property_value("number", 3)
    assert process_property_value(notion_data) == 3


def test_formula_date_empty():
    notion_data = mock_formula_property_value("date", None)
    assert process_property_value(notion_data) is None


def test_formula_date_start_only():
    notion_data = mock_formula_property_value("date", {"start": "2022-05-11"})
    assert process_property_value(notion_data) == "2022-05-11"


def test_relationship_empty():
    notion_data = mock_property_value("relation", [])
    assert process_property_value(notion_data) == []


def test_relationship_single():
    item = mock_relation_value()
    notion_data = mock_property_value("relation", [item])
    assert process_property_value(notion_data) == [item["id"]]


def test_relationship_double():
    item1 = mock_relation_value()
    item2 = mock_relation_value()
    notion_data = mock_property_value("relation", [item1, item2])
    assert process_property_value(notion_data) == [item1["id"], item2["id"]]


def test_rollup_date_empty():
    notion_data = mock_rollup_property_value("date", None)
    assert process_property_value(notion_data) is None


# TODO: Add some more tests for different types of rollups, including arrays of
# different types of values


@pytest.mark.xfail
def test_rollup_date_array():
    notion_data = mock_rollup_property_value(
        "array",
        [
            {"start": "2022-05-11"},
            {"start": "2022-05-12"},
            {"start": "2022-05-13"},
        ],
    )
    assert process_property_value(notion_data) == [
        "2022-05-11",
        "2022-05-12",
        "2022-05-12",
    ]



import os
import sys
import logging
import argparse

from n2y.logger import logger, HANDLER
from n2y.notion import Client
from n2y.page import Page
from n2y.database import Database
from n2y.mentions import PageMention
from n2y.errors import UseNextClass
from n2y.blocks import LinkToPageBlock
from n2y.utils import id_from_share_link


plugin_key = "audit"


class ReportingPageMention(PageMention):
    def __init__(self, client, notion_data, plain_text, block=None):
        if block is None:
            raise UseNextClass()
        super().__init__(client, notion_data, plain_text, block)
        if plugin_key not in block.page.plugin_data:
            block.page.plugin_data[plugin_key] = []
        block.page.plugin_data[plugin_key].append(
            {
                "block_url": block.notion_url,
                "linked_page_id": self.notion_page_id,
                "type": "page mention",
            }
        )


class ReportingLinkToPageBlock(LinkToPageBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        if plugin_key not in page.plugin_data:
            page.plugin_data[plugin_key] = []
        page.plugin_data[plugin_key].append(
            {
                "block_url": self.notion_url,
                "linked_page_id": self.linked_node_id,
                "type": "link to page",
            }
        )


def cli_main():
    args = sys.argv[1:]
    access_token = os.environ.get("NOTION_ACCESS_TOKEN", None)
    sys.exit(main(args, access_token))


def main(raw_args, access_token):
    parser = argparse.ArgumentParser(
        description="Audit a set of Notion pages for external links",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("object_id", help="The id or url for a Notion database or page")
    parser.add_argument(
        "--verbosity",
        "-v",
        default="INFO",
        help="Level to set the root logging module to",
    )
    parser.add_argument(
        "--logging-format",
        default="%(asctime)s - %(levelname)s - n2y.%(module)s: %(message)s",
        help="Default format used when logging",
    )

    args = parser.parse_args(raw_args)

    logging_level = logging.__dict__[args.verbosity]
    new_formatter = logging.Formatter(args.logging_format)
    logger.setLevel(logging_level)
    HANDLER.setFormatter(new_formatter)

    if access_token is None:
        logger.critical("No NOTION_ACCESS_TOKEN environment variable is set")
        return 1

    object_id = id_from_share_link(args.object_id)

    # TODO: handle synced blocks (need to upgrade the API version first so we
    # can go from synced-block to primary synced block then up to the page;
    # going up to the page from the other synced block isn't possible in the
    # 2022-02-22 API version)

    # TODO: handle plain old links to other pages
    # TODO: handle at-mentions outside of blocks
    plugins = {
        "mentions": {
            "page": ReportingPageMention,
        },
        "blocks": {
            "link_to_page": ReportingLinkToPageBlock,
        },
    }

    client = Client(access_token)
    client.load_plugin(plugins)

    node = client.get_page_or_database(object_id)

    if node is None:
        msg = (
            "Unable to find database or page with id %s. "
            "Perhaps its not shared with the integration?"
        )
        logger.error(msg, object_id)
        return 2

    references = {}
    audit_node(node, references, 0)
    external_references = exclude_internal_references(references)
    print_references(client, external_references)
    if any(len(l) for l in external_references.values()):
        return 3
    else:
        return 0


def exclude_internal_references(references):
    external_references = {}
    for page_id, links in references.items():
        external_links = [l for l in links if l["linked_page_id"] not in references]
        external_references[page_id] = external_links
    return external_references


def print_references(client, references):
    for page_id, links in references.items():
        if not links:
            continue
        page = client.get_page(page_id)
        print(page.title.to_plain_text())
        for link in links:
            print("  " + link["block_url"])


def audit_node(node, references, depth):
    if isinstance(node, Database):
        audit_database(node, references, depth)
    elif isinstance(node, Page):
        audit_page(node, references, depth)


def audit_database(database, references, depth):
    logger.info("%sDatabase %s", " " * depth, database.title.to_plain_text())
    for page in database.children:
        audit_page(page, references, depth + 1)


def audit_page(page, references, depth):
    logger.info("%sAuditing %s", " " * depth, page.title.to_plain_text())
    assert page.notion_id not in references  # expect that each page is visited once
    page.block  # load all of the blocks
    references[page.notion_id] = page.plugin_data.get(plugin_key, [])
    for node in page.children:
        audit_node(node, references, depth + 1)


if __name__ == "__main__":
    cli_main()



from n2y.logger import logger
from n2y.utils import fromisoformat, sanitize_filename


class Database:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating database")
        self.client = client

        self.notion_data = notion_data
        self.notion_id = notion_data["id"]
        self.created_time = fromisoformat(notion_data["created_time"])
        self.created_by = client.wrap_notion_user(notion_data["created_by"])
        self.last_edited_time = fromisoformat(notion_data["last_edited_time"])
        self.last_edited_by = client.wrap_notion_user(notion_data["last_edited_by"])
        self.title = client.wrap_notion_rich_text_array(notion_data["title"])
        self.icon = self._init_icon(notion_data["icon"])
        self.cover = notion_data["cover"] and client.wrap_notion_file(
            notion_data["cover"]
        )
        self.archived = notion_data["archived"]
        self.schema = {
            k: client.wrap_notion_property(p)
            for k, p in notion_data["properties"].items()
        }
        self.notion_parent = notion_data["parent"]
        self.notion_url = notion_data["url"]
        self.archived = notion_data["archived"]

        self._children = None
        self._filtered_children = {}

    @property
    def filename(self):
        return sanitize_filename(self.title.to_plain_text())

    @property
    def children(self):
        if self._children is None:
            self._children = self.client.get_database_pages(self.notion_id)
        return self._children

    def children_filtered(self, filter, sort=None):
        if filter is not None:
            tupled_filter = self._tuplize(filter)
            tupled_sort = self._tuplize(sort)
            if tupled_filter not in self._filtered_children:
                self._filtered_children[tupled_filter] = {}
            if tupled_sort not in self._filtered_children[tupled_filter]:
                self._filtered_children[tupled_filter][
                    tupled_sort
                ] = self.client.get_database_pages(self.notion_id, filter, sort)
            children = self._filtered_children[tupled_filter][tupled_sort]
        else:
            children = self.children
        return children

    def _tuplize(self, item):
        if callable(getattr(item, "items", None)):
            return tuple((key, self._tuplize(val)) for (key, val) in item.items())
        elif hasattr(item, "__iter__") and not isinstance(item, str):
            return tuple(self._tuplize(i) for i in item)
        else:
            return (item,)

    def _init_icon(self, icon_notion_data):
        """
        The icon property is unique in that it can be either an emoji or a file.
        """
        if icon_notion_data is None:
            return None
        elif icon_notion_data["type"] == "emoji":
            return self.client.wrap_notion_emoji(icon_notion_data)
        else:
            return self.client.wrap_notion_file(icon_notion_data)

    @property
    def parent(self):
        """
        The parent of a database can be a page, another database, a block, or
        it can be at the top level of workspace.
        """
        type = self.notion_parent["type"]
        if type == "workspace":
            return None

        if type == "page_id":
            return self.client.get_page(self.notion_parent["page_id"])

        if type == "database_id":
            return self.client.get_database(self.notion_parent["database_id"])

        if type == "block_id":
            return self.client.get_block(self.notion_parent["block_id"], page=None)

    def to_pandoc(self):
        return self.block.to_pandoc()



import importlib.util
import json
from os import makedirs, path
from urllib.parse import urljoin, urlparse

import requests

from n2y.blocks import DEFAULT_BLOCKS
from n2y.comment import Comment
from n2y.config import merge_default_config
from n2y.database import Database
from n2y.emoji import Emoji
from n2y.errors import (
    APIErrorCode,
    APIResponseError,
    HTTPResponseError,
    ObjectNotFound,
    PluginError,
    UseNextClass,
    is_api_error_code,
)
from n2y.file import File
from n2y.logger import logger
from n2y.mentions import DEFAULT_MENTIONS
from n2y.notion_mocks import mock_rich_text_array
from n2y.page import Page
from n2y.properties import DEFAULT_PROPERTIES
from n2y.property_values import DEFAULT_PROPERTY_VALUES
from n2y.rich_text import DEFAULT_RICH_TEXTS, RichTextArray
from n2y.user import User
from n2y.utils import retry_api_call, sanitize_filename, strip_hyphens

# TODO: Rename this file `client.py`

DEFAULT_NOTION_CLASSES = {
    "page": Page,
    "database": Database,
    "blocks": DEFAULT_BLOCKS,
    "properties": DEFAULT_PROPERTIES,
    "property_values": DEFAULT_PROPERTY_VALUES,
    "user": User,
    "file": File,
    "emoji": Emoji,
    "rich_text_array": RichTextArray,
    "rich_texts": DEFAULT_RICH_TEXTS,
    "mentions": DEFAULT_MENTIONS,
    "comment": Comment,
}


class Client:
    """
    An instance of the client class has a few purposes:

    1. To retrieve data from Notion
    2. To determine what classes to use to wrap this notion data
    3. To act as a shared global store for all of the objects that are pulled
       from Notion.

    In particular there is a cache of all pages and databases which ensure that
    the database class and page class are instantiated exactly once for each
    page or database.
    """

    def __init__(
        self,
        access_token,
        media_root=".",
        media_url="",
        plugins=None,
        export_defaults=None,
    ):
        self.access_token = access_token
        self.media_root = media_root
        self.media_url = media_url
        self.export_defaults = export_defaults or merge_default_config({})

        self.base_url = "https://api.notion.com/v1/"
        self.headers = {
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
            "Notion-Version": "2022-06-28",
        }

        self.databases_cache = {}
        self.pages_cache = {}
        self.users_cache = {}

        self.load_plugins(plugins)
        self.plugin_data = {}

    def get_default_classes(self):
        notion_classes = {}
        for notion_object, object_types in DEFAULT_NOTION_CLASSES.items():
            if type(object_types) is dict:
                notion_classes[notion_object] = {k: [v] for k, v in object_types.items()}
            else:
                notion_classes[notion_object] = [object_types]
        return notion_classes

    def load_plugins(self, plugins):
        self.notion_classes = self.get_default_classes()
        if plugins is not None:
            for plugin in plugins:
                plugin_module = importlib.import_module(plugin)
                try:
                    self.load_plugin(plugin_module.notion_classes)
                except PluginError as err:
                    logger.error('Error loading plugin "%s": %s', plugin, err)
                    raise

    def load_plugin(self, notion_classes):
        for notion_object, object_types in notion_classes.items():
            if notion_object in self.notion_classes:
                default_object_types = DEFAULT_NOTION_CLASSES[notion_object]
                self._override_notion_classes(
                    notion_object, object_types, default_object_types
                )
            else:
                raise PluginError(f'Invalid notion object "{notion_object}"')

    def _override_notion_classes(self, notion_object, object_types, default_object_types):
        # E.g., there are many types of notion blocks but only one type of notion page.
        notion_object_has_types = isinstance(default_object_types, dict)

        if notion_object_has_types and isinstance(object_types, dict):
            for object_type, plugin_class in object_types.items():
                self._organize_notion_classes(
                    default_object_types, notion_object, object_type, plugin_class
                )
        elif notion_object_has_types and not isinstance(object_types, dict):
            raise PluginError(
                f'Expecting dict for "{notion_object}", found "{type(object_types)}"'
            )
        else:
            plugin_class = object_types
            base_class = default_object_types
            if issubclass(plugin_class, base_class):
                self.notion_classes[notion_object].append(plugin_class)
            else:
                raise PluginError(
                    f'Cannot use "{plugin_class.__name__}", as it doesn\'t '
                    f'override the base class "{base_class.__name__}"',
                )

    def _organize_notion_classes(
        self, default_object_types, notion_object, object_type, plugin_class
    ):
        if object_type in default_object_types:
            class_being_replaced = default_object_types[object_type]
            # assumes all of the default classes have a single parent class
            base_class = class_being_replaced.__bases__[0]
            if issubclass(plugin_class, base_class):
                self.notion_classes[notion_object][object_type].append(plugin_class)
            else:
                raise PluginError(
                    f'Cannot use "{plugin_class.__name__}", as it doesn\'t '
                    f'override the base class "{base_class.__name__}"',
                )
        else:
            raise PluginError(f'Invalid type "{object_type}" for "{notion_object}"')

    def get_class_list(self, notion_object, object_type=None):
        if object_type is None:
            return self.notion_classes[notion_object]
        try:
            return self.notion_classes[notion_object][object_type]
        except KeyError:
            raise NotImplementedError(
                f'Unknown "{notion_object}" class of type "{object_type}"'
            )

    def instantiate_class(self, notion_object, object_type, *args, **kwargs):
        class_list = self.get_class_list(notion_object, object_type)
        for cls in reversed(class_list):
            try:
                return cls(*args, **kwargs)
            except UseNextClass:
                logger.debug("Skipping %s due to UseNextClass exception", cls.__name__)

    def _wrap_notion_page(self, notion_data):
        """
        Wrap notion page data in the appropriate class. If the page class has
        already been created, then ignore the new notion data and just return
        the existing version.

        Note that it may seem redundant to check the cache here, since it's also
        checked in `get_page`, however unlike databases which are only ever
        retrieved via `get_database`, pages can be retrieved either through
        `get_page` or through `get_database_pages`. Thus, it's possible that a
        database page be first retrieved individually and then retrieved via the
        full database. In this case, it's unfortunate (but unavoidable) that we
        retrieved the data from Notion twice, but even so we don't want to
        replace our existing page instance, along with it's content or other
        state that has been added to it.
        """
        page_in_cache = notion_data["id"] in self.pages_cache
        if page_in_cache and self.class_is_in_use(
            # Need to check that the page in the client cache was instantiated using
            # the currently favored page class. Otherwise, plugins set for one export
            # will be used in another or a plugin will be set but not used.
            # As we currently use the `jinjarenderpage` plugin for all pages,
            # this check is most likely uneccessary at this point.
            self.pages_cache[notion_data["id"]],
            "page",
        ):
            return self.pages_cache[notion_data["id"]]
        else:
            page = self.instantiate_class("page", None, self, notion_data)
            not page_in_cache or logger.warning(
                f"page in cache overwritten at key \"{notion_data['id']}\""
            )
            self.pages_cache[page.notion_id] = page
            return page

    def _wrap_notion_database(self, notion_data):
        return self.instantiate_class("database", None, self, notion_data)

    def wrap_notion_block(self, notion_data, page, get_children):
        return self.instantiate_class(
            "blocks",
            notion_data["type"],
            self,
            notion_data,
            page,
            get_children,
        )

    def wrap_notion_user(self, notion_data):
        """
        Retrieve the user if its not in the cache.
        """
        user_id = notion_data["id"]
        if user_id in self.users_cache and self.class_is_in_use(
            self.users_cache[user_id], "user"
        ):
            user = self.users_cache[user_id]
        else:
            if [key for key in notion_data] == ["object", "id"]:
                try:
                    notion_data = self.get_notion_user(user_id)
                except ObjectNotFound:
                    pass
            user = self.instantiate_class("user", None, self, notion_data)
            self.users_cache[user_id] = user
        return user

    def wrap_notion_file(self, notion_data):
        return self.instantiate_class("file", None, self, notion_data)

    def wrap_notion_emoji(self, notion_data):
        return self.instantiate_class("emoji", None, self, notion_data)

    def wrap_notion_rich_text_array(self, notion_data, block=None):
        return self.instantiate_class("rich_text_array", None, self, notion_data, block)

    def wrap_notion_rich_text(self, notion_data, block=None):
        return self.instantiate_class(
            "rich_texts", notion_data["type"], self, notion_data, block
        )

    def wrap_notion_mention(self, notion_data, plain_text, block=None):
        # here we pass in the plain_text to avoid the need to query the page
        # just to get its title
        return self.instantiate_class(
            "mentions",
            notion_data["type"],
            self,
            notion_data,
            plain_text,
            block,
        )

    def wrap_notion_property(self, notion_data):
        return self.instantiate_class(
            "properties", notion_data["type"], self, notion_data
        )

    def wrap_notion_property_value(self, notion_data, page):
        return self.instantiate_class(
            "property_values", notion_data["type"], self, notion_data, page
        )

    def get_page_or_database(self, object_id):
        """
        First attempt to get the page corresponding with the object id; if
        the page doesn't exist, then attempt to retrieve the database. This
        trial-and-error is necessary because the API doesn't provide a means to
        determining what type of object corresponds with an ID and we don't want
        to make the user indicate if they are pulling down a database or a page.
        """
        page = self.get_page(object_id)
        if page is not None:
            return page
        else:
            return self.get_database(object_id)

    def get_database(self, database_id):
        """
        Retrieve the database (but not it's pages) if its not in the cache. Even
        if it is in the cache.
        """
        if database_id in self.databases_cache and self.class_is_in_use(
            self.databases_cache[database_id], "database"
        ):
            database = self.databases_cache[database_id]
        else:
            try:
                notion_database = self.get_notion_database(database_id)
                database = self._wrap_notion_database(notion_database)
            except ObjectNotFound:
                database = None
            self.databases_cache[database_id] = database
        return database

    def get_notion_database(self, database_id):
        return self._get_url(f"{self.base_url}databases/{database_id}")

    def get_database_pages(self, database_id, filter=None, sorts=None):
        notion_pages = self.get_database_notion_pages(database_id, filter, sorts)
        return [self._wrap_notion_page(np) for np in notion_pages]

    def get_database_notion_pages(self, database_id, filter, sorts):
        url = f"{self.base_url}databases/{database_id}/query"
        request_data = {}
        if filter:
            request_data["filter"] = filter
        if sorts:
            request_data["sorts"] = sorts
        return self._paginated_request(self._post_url, url, request_data)

    def get_page(self, page_id):
        """
        Retrieve the page if its not in the cache.
        """
        if page_id in self.pages_cache:
            page = self.pages_cache[page_id]
            if not self.class_is_in_use(page, "page"):
                page = self.instantiate_class("page", None, self, page.notion_data)
        else:
            try:
                notion_page = self.get_notion_page(page_id)
            except ObjectNotFound:
                self.pages_cache[page_id] = None
                return
            # _wrap_notion_page will add the page to the cache
            page = self._wrap_notion_page(notion_page)
        return page

    def get_notion_page(self, page_id):
        return self._get_url(f"{self.base_url}pages/{page_id}")

    def get_notion_user(self, user_id):
        return self._get_url(f"{self.base_url}users/{user_id}")

    def get_block(self, block_id, page, get_children=True):
        notion_block = self.get_notion_block(block_id)
        return self.wrap_notion_block(notion_block, page, get_children)

    def get_notion_block(self, block_id):
        url = f"{self.base_url}blocks/{block_id}"
        return self._get_url(url)

    def get_child_blocks(self, block_id, page, get_children):
        child_notion_blocks = self.get_child_notion_blocks(block_id)
        return [
            self.wrap_notion_block(b, page, get_children) for b in child_notion_blocks
        ]

    def get_child_notion_blocks(self, block_id):
        url = f"{self.base_url}blocks/{block_id}/children"
        return self._paginated_request(self._get_url, url, {})

    def get_comments(self, block_id):
        url = f"{self.base_url}comments"
        comments = self._paginated_request(self._get_url, url, {"block_id": block_id})
        return [self.wrap_notion_comment(nd) for nd in comments]

    def wrap_notion_comment(self, notion_data):
        return self.instantiate_class("comment", None, self, notion_data)

    def get_page_property(self, page_id, property_id):
        notion_property = self.get_notion_page_property(page_id, property_id)
        return self.wrap_notion_property(notion_property)

    def get_notion_page_property(self, page_id, property_id):
        url = f"{self.base_url}pages/{page_id}/properties/{property_id}"
        return self._get_url(url)

    def create_notion_page(self, page_data):
        creation_url = f"{self.base_url}pages"
        return self._post_url(creation_url, page_data)

    @retry_api_call
    def _get_url(self, url, params=None, stream=False):
        if not stream and params is None:
            params = {}
        response = requests.get(
            url,
            params=params,
            stream=stream,
            headers=self.headers if not stream else None,
        )
        return self._parse_response(response, stream)

    @retry_api_call
    def _post_url(self, url, data=None):
        if data is None:
            data = {}
        response = requests.post(url, headers=self.headers, json=data)
        return self._parse_response(response)

    @retry_api_call
    def _delete_url(self, url):
        response = requests.delete(
            url, headers={k: v for k, v in self.headers.items() if k != "Content-Type"}
        )
        return self._parse_response(response)

    @retry_api_call
    def _patch_url(self, url, data=None):
        if data is None:
            data = {}
        response = requests.patch(url, headers=self.headers, json=data)
        return self._parse_response(response)

    def _paginated_request(self, request_method, url, initial_params):
        params = initial_params
        results = []
        while True:
            data = request_method(url, params)
            results.extend(data["results"])
            if not data["has_more"]:
                return results
            else:
                params["start_cursor"] = data["next_cursor"]

    def _parse_response(self, response, stream=False):
        """Taken from https://github.com/ramnes/notion-sdk-py"""
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            try:
                body = error.response.json()
                code = body.get("code")
            except json.JSONDecodeError:
                code = None
            if code == APIErrorCode.ObjectNotFound:
                raise ObjectNotFound(response, body["message"])
            elif code and is_api_error_code(code):
                raise APIResponseError(response, body["message"], code)
            raise HTTPResponseError(error.response)
        return response.json() if not stream else response.content

    def download_file(self, url, page, block_id):
        """
        Download a file from a given URL into the MEDIA_ROOT.

        Preserve the file extension from the URL, but use the
        id of the block followed by an md5 hash.
        """
        url_path = path.basename(urlparse(url).path)
        _, extension = path.splitext(url_path)
        content = self._get_url(url, stream=True)
        return self.save_file(content, page, extension, block_id)

    def save_file(self, content, page, extension, block_id):
        block_id_chars = strip_hyphens(block_id)
        page_title = sanitize_filename(page.title.to_plain_text())
        relative_filepath = f"{page_title}-{block_id_chars[:11]}{extension}"
        full_filepath = path.join(self.media_root, relative_filepath)
        makedirs(self.media_root, exist_ok=True)
        with open(full_filepath, "wb") as temp_file:
            temp_file.write(content)
        return urljoin(self.media_url, relative_filepath)

    def copy_notion_database_children(self, children, destination):
        """
        copy the notion children (`children`) of one notion database to another (`destination`)
        """
        bad_keys = [
            "id",
            "url",
            "parent",
            "created_by",
            "created_time",
            "last_edited_by",
            "last_edited_time",
        ]
        db_children = [
            {key: value for (key, value) in child.items() if key not in bad_keys}
            for child in children
        ]
        for page in db_children:
            page["parent"] = {"type": "database_id", "database_id": destination["id"]}
            for key in page["properties"].keys():
                page["properties"][key] = self._edit_notion_child_property(
                    page["properties"][key]
                )
            self.create_notion_page(page)

    def _edit_notion_child_property(self, prop):
        del prop["id"]
        prop_type = prop["type"]
        del prop["type"]
        prop_type_info = prop[prop_type]
        if isinstance(prop_type_info, dict):
            if "id" in prop_type_info:
                del prop_type_info["id"]
        elif isinstance(prop_type_info, list) and prop_type != "relation":
            for item in prop_type_info:
                if "id" in item:
                    del item["id"]
        return prop

    def _notion_block_type_is(self, block, type):
        return "type" in block and block["type"] == type

    def _notion_block_object_is(self, block, object):
        return "object" in block and block["object"] == object

    def append_child_notion_blocks(self, block_id, children):
        """
        Appends each datapoint of a list of notion_data as children to the block specified by id.
        Please note that not all block types are allowed to have children, so this method only works
        for those that are.
        """
        previous_i = 0
        children_appended = []
        parent = self.get_page_or_database(block_id) or self.get_block(block_id, None)
        parent_type = parent.notion_data["object"]

        for i, child in enumerate(children):
            if self._notion_block_object_is(
                child, "database"
            ) or self._notion_block_type_is(child, "child_database"):
                if previous_i != i:
                    children_appended = self._append_blocks(
                        block_id, children, children_appended, previous_i, i
                    )
                child_database = self._copy_notion_database_child_database(
                    parent, parent_type, child
                )
                children_appended.append(child_database)
                previous_i = i + 1
            elif self._notion_block_object_is(
                child, "page"
            ) or self._notion_block_type_is(child, "child_page"):
                if previous_i != i:
                    children_appended = self._append_blocks(
                        block_id, children, children_appended, previous_i, i
                    )
                child_page = self._copy_notion_database_child_page(
                    parent, parent_type, child
                )
                children_appended.append(child_page)
                previous_i = i + 1
            elif i == len(children) - 1:
                children_appended = self._append_blocks(
                    block_id, children, children_appended, previous_i, len(children)
                )
        return children_appended

    def _append_blocks(
        self, block_id, full_child_data_list, appension_history_list, i1=0, i2=0
    ):
        max_new_blocks = 100
        if i1 < i2:
            child_data_list = full_child_data_list[i1:i2]
            length = len(child_data_list)
            for i in range(0, length, max_new_blocks):
                portion_index_stop = i + max_new_blocks
                if portion_index_stop < length:
                    portion = child_data_list[i:portion_index_stop]
                else:
                    portion = child_data_list[i:]
                appension_return = self._patch_url(
                    f"{self.base_url}blocks/{block_id}/children", {"children": portion}
                )
                appension_history_list.extend(appension_return["results"])
        return appension_history_list

    def _copy_notion_database_child_page(self, parent, parent_type, child_notion_data):
        bad_keys = [
            "last_edited_by",
            "created_time",
            "last_edited_time",
            "created_by",
            "request_id",
        ]
        if parent_type == "block":
            logger.warning(
                "Skipping page with block type parent as "
                "appension is currently unsupported by Notion API"
            )
            child_page = {}
        else:
            page = self.get_page(child_notion_data["id"])
            child_page_data = {
                key: value
                for (key, value) in page.notion_data.items()
                if key not in bad_keys
            }
            child_page_data["parent"] = {
                "type": f"{parent_type}_id",
                f"{parent_type}_id": parent.notion_id,
            }
            child_page = self.create_notion_page(child_page_data)
        return child_page

    def _copy_notion_database_child_database(
        self, parent, parent_type, child_notion_data
    ):
        bad_keys = [
            "last_edited_by",
            "created_time",
            "last_edited_time",
            "created_by",
            "request_id",
        ]
        if parent_type == "block":
            logger.warning(
                "Skipping database with block type parent as "
                "appension is currently unsupported by Notion API"
            )
            child_database = {}
        else:
            database = self.get_database(child_notion_data["id"])
            child_database_data = {
                key: value
                for (key, value) in database.notion_data.items()
                if key not in bad_keys
            }
            child_database_data["parent"] = {
                "type": f"{parent_type}_id",
                f"{parent_type}_id": parent.notion_id,
            }
            child_database = self.create_notion_database(child_database_data)
            if database.children:
                notion_children = [child.notion_data for child in database.children]
                self.copy_notion_database_children(notion_children, child_database)
        return child_database

    def create_notion_comment(self, page_id, text_blocks_descriptors):
        data = {
            "rich_text": mock_rich_text_array(text_blocks_descriptors),
            "parent": {
                "type": "page_id",
                "page_id": page_id,
            },
        }
        return self._post_url(f"{self.base_url}comments", data)

    def create_notion_database(self, notion_data):
        return self._post_url(f"{self.base_url}databases", notion_data)

    def delete_notion_block(self, notion_block):
        return self._delete_url(f"{self.base_url}blocks/{notion_block['id']}")

    def class_is_in_use(self, class_object, cls):
        """
        Checks if the given class object has been instantiated
        with the currently favored class type for that class.
        """
        if class_object is None or type(class_object) is self.notion_classes[cls][-1]:
            return True
        return False



from n2y.blocks import NoopBlock


class RemoveCalloutBlock(NoopBlock):
    """
    Completely removes callout blocks from the generated content.

    Useful, e.g., if you want callout blocks to contain help text that shouldn't
    be included in generated documents.
    """

    ...


notion_classes = {
    "blocks": {
        "callout": RemoveCalloutBlock,
    }
}



"""
A setuptools based setup module.
"""

from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

description = "Notion to YAML"

setup(
    name="n2y",
    version="0.8.2",
    description=description,
    long_description=description,
    long_description_content_type="text/x-rst",
    url="https://github.com/innolitics/n2y",
    author="Innolitics, LLC",
    author_email="info@innolitics.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="notion documentation yaml markdown",
    packages=find_packages(exclude=["tests"]),
    install_requires=["pyyaml", "requests", "pandoc", "jinja2"],
    extras_require={
        "dev": [
            "pytest",
            "flake8",
            "check-manifest",
            "requests-cache",
        ]
    },
    include_package_data=True,
    package_data={"n2y": ["data/mermaid_err.png"]},
    entry_points={
        "console_scripts": [
            "n2y = n2y.main:cli_main",
            "n2yaudit = n2y.audit:cli_main",
        ]
    },
)



import os
import sys
import yaml
import logging
import argparse
from importlib.metadata import version

from n2y.logger import logger
from n2y.notion import Client
from n2y.utils import share_link_from_id
from n2y.config import load_config, merge_default_config
from n2y.export import export_page, database_to_yaml, database_to_files, write_document


def cli_main():
    args = sys.argv[1:]
    access_token = os.environ.get("NOTION_ACCESS_TOKEN", None)
    n2y_cache = os.environ.get("N2Y_CACHE", None)
    sys.exit(main(args, access_token, n2y_cache))


def main(raw_args, access_token, n2y_cache=None):
    parser = argparse.ArgumentParser(
        description="Move data from Notion into YAML/markdown",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("config", help="The path to the config file")
    parser.add_argument(
        "--verbosity",
        "-v",
        default="INFO",
        help="Level to set the root logging module to",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=version("n2y"),
        help="The version of n2y installed",
    )

    args = parser.parse_args(raw_args)

    logging_level = logging.__dict__[args.verbosity]
    logger.setLevel(logging_level)

    if n2y_cache is not None:
        try:
            import requests_cache

            requests_cache.install_cache(n2y_cache, backend="sqlite", expire_after=-1)
            logger.info("Using cache at %s", n2y_cache)
        except ImportError:
            logger.warning(
                "The requests_cache module is not installed. Ignoring N2Y_CACHE %s",
                n2y_cache,
            )

    if access_token is None:
        logger.critical("No NOTION_ACCESS_TOKEN environment variable is set")
        return 1

    config = load_config(args.config)
    if config is None:
        return 2
    export_defaults = merge_default_config(config.get("export_defaults", {}))

    client = Client(
        access_token,
        config["media_root"],
        config["media_url"],
        export_defaults=export_defaults,
    )

    error_occurred = False
    for export in config["exports"]:
        logger.info("Exporting to %s", export["output"])
        client.load_plugins(export["plugins"])
        export_completed = _export_node_from_config(client, export)
        if not export_completed:
            error_occurred = True
    return 0 if not error_occurred else 3


def _export_node_from_config(client, export):
    node_type = export["node_type"]
    if node_type == "page":
        page = client.get_page(export["id"])
        if page is None:
            msg = (
                "Unable to find page with id '%s' (%s). "
                "Perhaps the integration doesn't have permission to access this page?"
            )
            logger.error(msg, export["id"], share_link_from_id(export["id"]))
            return False
        document = export_page(
            page,
            export["pandoc_format"],
            export["pandoc_options"],
            export["yaml_front_matter"],
            export["id_property"],
            export["url_property"],
            export["property_map"],
        )
        write_document(document, export["output"])
    else:
        database = client.get_database(export["id"])
        if database is None:
            msg = (
                "Unable to find database with id '%s' (%s). "
                "Perhaps the integration doesn't have permission to access this database?"
            )
            logger.error(msg, export["id"], share_link_from_id(export["id"]))
            return False
        if node_type == "database_as_yaml":
            result = database_to_yaml(
                database=database,
                pandoc_format=export["pandoc_format"],
                pandoc_options=export["pandoc_options"],
                id_property=export["id_property"],
                url_property=export["url_property"],
                content_property=export["content_property"],
                notion_filter=export["notion_filter"],
                notion_sorts=export["notion_sorts"],
                property_map=export["property_map"],
            )
            result = yaml.dump(result, sort_keys=False)
            write_document(result, export["output"])
        elif node_type == "database_as_files":
            database_to_files(
                database=database,
                directory=export["output"],
                pandoc_format=export["pandoc_format"],
                pandoc_options=export["pandoc_options"],
                yaml_front_matter=export["yaml_front_matter"],
                filename_template=export["filename_template"],
                notion_filter=export["notion_filter"],
                notion_sorts=export["notion_sorts"],
                id_property=export["id_property"],
                url_property=export["url_property"],
                property_map=export["property_map"],
            )
        else:
            logger.error("Unknown node_type '%s'", node_type)
            return False
    return True


if __name__ == "__main__":
    cli_main()



from unittest.mock import patch

import pytest

from n2y import notion
from n2y.export import _page_filename, _page_properties
from n2y.notion_mocks import mock_page, mock_rich_text_property_value, mock_user
from n2y.user import User

pdf = "T.pdf"


@pytest.fixture
@patch("n2y.notion.Client.wrap_notion_user")
def page(wrap_notion_user):
    property_value = mock_rich_text_property_value("P")
    notion_page = mock_page(title="T", extra_properties={"property": property_value})
    client = notion.Client("")
    wrap_notion_user.return_value = User(client, mock_user())
    return client._wrap_notion_page(notion_page)


def test_page_properties_basic(page):
    properties = _page_properties(page)
    assert properties == {"title": "T", "property": "P\n"}


def test_page_properties_id(page):
    properties = _page_properties(page, id_property="id")
    assert "id" in properties


def test_page_properties_url(page):
    properties = _page_properties(page, url_property="url")
    assert "url" in properties


def test_page_properties_mapping(page):
    properties = _page_properties(page, property_map={"property": "p"})
    assert properties == {"title": "T", "p": "P\n"}
    properties = _page_properties(page, property_map={"property": None})
    assert properties == {"title": "T"}


def test_page_filename_no_template(page):
    assert _page_filename(page, "pdf") == pdf
    assert _page_filename(page, "pdf+extra") == pdf
    assert _page_filename(page, "pdf-extra") == pdf
    assert _page_filename(page, "pdf-extra+other") == pdf


def test_page_filename_template(page):
    assert _page_filename(page, "pdf", "{property}.p") == "P.p"
    assert _page_filename(page, "pdf", "{TITLE}.p") == "T.p"
    assert _page_filename(page, "pdf", "{TITLE}-{property}.p") == "T-P.p"


def test_page_filename_template_malformed(page):
    assert _page_filename(page, "pdf", "{missing}.p") == pdf



from n2y.logger import logger
from n2y.blocks import LinkToPageBlock


class ExpandingLinkToPageBlock(LinkToPageBlock):
    """
    Replace page link with the content of the linked-to page.

    """

    def to_pandoc(self):
        assert self.linked_node_id is not None

        if self.link_type == "page_id":
            page = self.client.get_page(self.linked_node_id)
            # The `page.block` refers to the ChildPageBlock in the page; we don't
            # want to call `to_pandoc` on it directly, since we don't want a
            # full pandoc document, but just the content that would have been in
            # that document.
            return page.block.children_to_pandoc()
        else:
            # TODO: Might be expanded to handle links to databases as well.
            logger.warning(
                "Links to databases (to:%s from:%s) not supported at this time.",
                self.linked_node_id,
                self.page.notion_id,
            )
            return None


notion_classes = {
    "blocks": {
        "link_to_page": ExpandingLinkToPageBlock,
    }
}



import uuid
from datetime import datetime, timezone

from n2y.utils import strip_hyphens


def mock_id():
    return str(uuid.uuid4())


def mock_user(**kwargs):
    return {"object": "user", "id": mock_id(), **kwargs}


def mock_person_user(name, email):
    return mock_user(name=name, type="person", person={"email": email})


def mock_rich_text_array(text_blocks_descriptors):
    if isinstance(text_blocks_descriptors, str):
        return [mock_rich_text(text_blocks_descriptors, [])]
    else:
        return [mock_rich_text(*desc) for desc in text_blocks_descriptors]


def mock_rich_text(
    text, annotations=None, href=None, mention=None, link=None, equation=None
):
    if annotations is None:
        annotations = []
    if mention:
        rich_text_type = "mention"
        content = mention
    elif equation:
        rich_text_type = "equation"
        content = equation
    else:
        rich_text_type = "text"
        content = {"content": text, "link": link}
    return {
        "type": rich_text_type,
        "annotations": mock_annotations(annotations),
        "plain_text": text,
        "href": href,
        rich_text_type: content,
    }


def mock_annotations(annotations=None):
    if annotations is None:
        annotations = []
    color_string = [a for a in annotations if "color" in a]
    if color_string:
        color = color_string[0].replace("color:", "")
    else:
        color = "default"
    return {
        "bold": True if "bold" in annotations else False,
        "italic": True if "italic" in annotations else False,
        "strikethrough": True if "strikethrough" in annotations else False,
        "underline": True if "underline" in annotations else False,
        "code": True if "code" in annotations else False,
        "color": color,
    }


def mock_user_mention():
    return {
        "type": "user",
        "user": mock_user(),
    }


def mock_page_mention():
    return {
        "type": "page",
        "page": {
            "id": mock_id(),
        },
    }


def mock_database_mention(id=None):
    return {
        "type": "database",
        "database": {
            "id": id or mock_id(),
        },
    }


def mock_block(block_type, content, has_children=False, **kwargs):
    created_by = mock_user()
    created_time = datetime.now(tz=timezone.utc).isoformat()
    notion_id = mock_id()
    return {
        "id": notion_id,
        "url": f"#{strip_hyphens(notion_id)}",
        "created_time": created_time,
        "created_by": created_by,
        "last_edited_time": created_time,
        "last_edited_by": created_by,
        "object": "block",
        "has_children": has_children,
        "archived": False,
        "type": block_type,
        block_type: content,
        **kwargs,
    }


def mock_paragraph_block(text_blocks_descriptors, **kwargs):
    return mock_block(
        "paragraph",
        {
            "color": "default",
            "rich_text": mock_rich_text_array(text_blocks_descriptors),
        },
        **kwargs,
    )


def mock_file(url):
    return {"url": url, "expiry_time": None}


def mock_property_value(property_value_type, content):
    return {
        "id": mock_id(),
        "type": property_value_type,
        property_value_type: content,
    }


def mock_rich_text_property_value(text_blocks_descriptors):
    return mock_property_value("rich_text", mock_rich_text_array(text_blocks_descriptors))


def mock_formula_property_value(formula_type, content):
    return mock_property_value(
        "formula",
        {
            "type": formula_type,
            formula_type: content,
        },
    )


def mock_rollup_property_value(rollup_type, content):
    return mock_property_value(
        "rollup",
        {
            "type": rollup_type,
            "function": "function",
            rollup_type: content,
        },
    )


def mock_select_option(name, **kwargs):
    return {
        "id": mock_id(),
        "name": name,
        "color": "green",
        **kwargs,
    }


def mock_relation_value():
    return {"id": mock_id()}


def mock_page(title="Mock Page", extra_properties=None):
    if extra_properties is None:
        extra_properties = {}
    user = mock_user()
    created_time = datetime.now(tz=timezone.utc).isoformat()
    notion_id = mock_id()
    hyphenated_title = title.replace(" ", "-")
    return {
        "object": "page",
        "id": notion_id,
        "created_time": created_time,
        "last_edited_time": created_time,
        "created_by": user,
        "last_edited_by": user,
        "cover": None,
        "icon": None,
        "parent": {"type": "page_id", "page_id": mock_id()},
        "archived": False,
        "properties": {
            "title": {
                "id": "title",
                "type": "title",
                "title": mock_rich_text_array(title),
            },
            **extra_properties,
        },
        "url": f"https://www.notion.so/{hyphenated_title}-{notion_id}",
    }


def mock_database(title="Mock Database", extra_properties=None):
    if extra_properties is None:
        extra_properties = {}
    hyphenated_title = title.replace(" ", "-")
    created_time = datetime.now(tz=timezone.utc).isoformat()
    notion_id = mock_id()
    user = mock_user()
    return {
        "object": "database",
        "id": notion_id,
        "cover": None,
        "icon": None,
        "created_time": created_time,
        "last_edited_time": created_time,
        "created_by": user,
        "last_edited_by": user,
        "title": mock_rich_text_array(title),
        "description": [],
        "is_inline": False,
        "archived": False,
        "properties": {
            "Name": {
                "id": "title",  # all title properties have id 'title'
                "name": "Name",
                "type": "title",
                "title": {},
            },
            **extra_properties,
        },
        "parent": {"type": "page_id", "page_id": mock_id()},
        "url": f"https://www.notion.so/{hyphenated_title}-{notion_id}",
    }


def mock_comment(text_blocks_descriptors: list[list[str | list[str]]], **kwargs) -> dict:
    date = datetime.now(tz=timezone.utc).isoformat()
    return {
        "id": mock_id(),
        "parent": mock_id(),
        "discussion_id": mock_id(),
        "created_time": date,
        "last_edited_time": date,
        "created_by": mock_user(),
        "rich_text": mock_rich_text_array(text_blocks_descriptors),
        **kwargs,
    }



from datetime import datetime

from n2y.logger import logger
from n2y.utils import fromisoformat, process_notion_date, processed_date_to_plain_text


class PropertyValue:
    def __init__(self, client, notion_data, page):
        self.client = client
        self.notion_property_id = notion_data.get(
            "id", None
        )  # will be none for rollup array values
        self.notion_type = notion_data["type"]
        self.page = page

    def to_value(self, pandoc_format=None):
        raise NotImplementedError()


class TitlePropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        # TODO: handle the case when there are more than 25 rich text items in the property
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data, page)
        self.rich_text = client.wrap_notion_rich_text_array(notion_data["title"])

    def to_value(self, _=None):
        # Notion allows styling of the title, however, in their UI they display
        # the title property without any styling. Thus, if you copy/paste styled
        # text into a title this styling can be hidden and can re-appear after
        # the document conversion. To avoid this surprise, we only generate
        # plain text here.
        return self.rich_text.to_plain_text()


class TextPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        # TODO: handle the case when there are more than 25 rich text items in the property
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data, page)
        self.rich_text = client.wrap_notion_rich_text_array(notion_data["rich_text"])

    def to_value(self, pandoc_format="gfm"):
        if pandoc_format is None:
            return self.rich_text.to_plain_text()
        else:
            return self.rich_text.to_value(pandoc_format)


class NumberPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.number = notion_data["number"]

    def to_value(self, _=None):
        return self.number


class StatusPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.status = notion_data["status"]

    def to_value(self, _=None):
        # Note: the Notion UI shouldn't allow you to have two statuses with the
        # same name nor should it allow no status at all
        return self.status


class SelectPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        notion_select = notion_data["select"]
        if notion_select is not None:
            self.notion_option_id = notion_select["id"]
            self.name = notion_select["name"]
            self.color = notion_select["color"]
        else:
            self.notion_option_id = None
            self.name = None
            self.color = None

    def to_value(self, _=None):
        # Note: the Notion UI shouldn't allow you to have two options with the
        # same name
        return self.name


class MultiSelectPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.options = [
            MultiSelectOption(self.client, no) for no in notion_data["multi_select"]
        ]

    def to_value(self, _=None):
        # Note: the Notion UI shouldn't allow you to have two options with the
        # same name
        return [o.name for o in self.options]


class MultiSelectOption:
    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option["id"]
        self.name = notion_option["name"]
        self.color = notion_option["color"]


class DatePropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        # TODO: handle timezones
        super().__init__(client, notion_data, page)
        self.value = process_notion_date(notion_data["date"])

    def to_value(self, _=None):
        return self.value

    def to_plain_text(self):
        return processed_date_to_plain_text(self.value)


class PeoplePropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.people = [client.wrap_notion_user(nu) for nu in notion_data["people"]]

    def to_value(self, _=None):
        return [u.to_value() for u in self.people]


class FilesPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.files = [client.wrap_notion_file(nf) for nf in notion_data["files"]]

    def to_value(self, _):
        return [f.to_value() for f in self.files]


class CheckboxPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.checkbox = notion_data["checkbox"]

    def to_value(self, _):
        return self.checkbox


class UrlPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.url = notion_data["url"]

    def to_value(self, _=None):
        return self.url


class EmailPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.email = notion_data["email"]

    def to_value(self, _=None):
        return self.email


class PhoneNumberPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.phone_number = notion_data["phone_number"]

    def to_value(self, _=None):
        return self.phone_number


class FormulaPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        # TODO: set other attributes
        notion_formula = notion_data["formula"]
        if notion_formula["type"] == "date":
            self.value = process_notion_date(notion_formula["date"])
        else:
            self.value = notion_formula[notion_formula["type"]]

    def to_value(self, _=None):
        return self.value


class RelationPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        if "has_more" not in notion_data or not notion_data["has_more"]:
            self.ids = [related["id"] for related in notion_data["relation"]]
        else:
            url = f"{client.base_url}pages/{page.notion_id}/properties/{self.notion_property_id}"
            self.ids = [
                r["relation"]["id"]
                for r in client._paginated_request(client._get_url, url, {})
            ]

    def to_value(self, _=None):
        return self.ids


class RollupPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        # TODO: handle the case when the rollup needs to be paginated
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data, page)
        notion_rollup = notion_data["rollup"]
        self.rollup_type = notion_rollup["type"]
        self.function = notion_rollup["function"]
        if self.rollup_type == "date":
            self.value = process_notion_date(notion_rollup["date"])
        elif self.rollup_type == "string":
            self.value = notion_rollup["string"]
        elif self.rollup_type == "number":
            self.value = notion_rollup["number"]
        elif self.rollup_type == "array":
            self.value = [
                self.client.wrap_notion_property_value(pv, page)
                for pv in notion_rollup["array"]
            ]
        else:
            logger.warning("Unhandled rollup type %s", notion_rollup["type"])
            self.value = notion_rollup[notion_rollup["type"]]
        # TODO: handle arrays of dates

    def to_value(self, pandoc_format="gfm"):
        if self.rollup_type == "date":
            return self.value
        elif self.rollup_type == "string":
            return self.value
        elif self.rollup_type == "number":
            return self.value
        elif self.rollup_type == "array":
            return [pv.to_value(pandoc_format) for pv in self.value]
        else:
            return self.value


class CreatedTimePropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.created_time = fromisoformat(notion_data["created_time"])

    def to_value(self, _=None):
        return datetime.isoformat(self.created_time)


class CreatedByPropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.created_by = client.wrap_notion_user(notion_data["created_by"])

    def to_value(self, _=None):
        return self.created_by.to_value()


class LastEditedTimePropertyValue(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.last_edited_time = fromisoformat(notion_data["last_edited_time"])

    def to_value(self, _=None):
        return datetime.isoformat(self.last_edited_time)


class LastEditedBy(PropertyValue):
    def __init__(self, client, notion_data, page):
        super().__init__(client, notion_data, page)
        self.last_edited_by = client.wrap_notion_user(notion_data["last_edited_by"])

    def to_value(self, _=None):
        return self.last_edited_by.to_value()


DEFAULT_PROPERTY_VALUES = {
    "title": TitlePropertyValue,
    "rich_text": TextPropertyValue,
    "number": NumberPropertyValue,
    "status": StatusPropertyValue,
    "select": SelectPropertyValue,
    "multi_select": MultiSelectPropertyValue,
    "date": DatePropertyValue,
    "people": PeoplePropertyValue,
    "files": FilesPropertyValue,
    "checkbox": CheckboxPropertyValue,
    "url": UrlPropertyValue,
    "email": EmailPropertyValue,
    "phone_number": PhoneNumberPropertyValue,
    "formula": FormulaPropertyValue,
    "relation": RelationPropertyValue,
    "rollup": RollupPropertyValue,
    "created_time": CreatedTimePropertyValue,
    "created_by": CreatedByPropertyValue,
    "last_edited_time": LastEditedTimePropertyValue,
    "last_edited_by": LastEditedBy,
}
